/*
************************************************************************
*
*   GLView.c - GL viewing parameters
*
*   Copyright (c) 1994
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : 96/03/02
*   Pathname of SCCS file     : /sgiext/molmol/sg/src/gl/SCCS/s.GLView.c
*   SCCS identification       : 1.5
*
************************************************************************
*/

#include "gl_view.h"

#include <math.h>
#include <gl/gl.h>
#include <gl/get.h>

#include <sg_get.h>
#include "gl_util.h"

/* there seems to be no clean way to get these values */
#define YMAXSTEREO 492
#define YOFFSET 532

static BOOL StereoSwitch = FALSE, StereoState = FALSE;
static long MonoMonitor;

static SgProjection Proj;
static Coord VX, VY, VZ;
static Angle ViewAng;
static Coord Near, Far;
static float VpX, VpY, VpW, VpH;
static float AspRat;

static Matrix IdentM = {
  {1.0, 0.0, 0.0, 0.0},
  {0.0, 1.0, 0.0, 0.0},
  {0.0, 0.0, 1.0, 0.0},
  {0.0, 0.0, 0.0, 1.0}
};

static Matrix ProjM;

void
SgGLSetStereo(BOOL onOff)
{
  if (onOff == StereoSwitch)
    return;

  if (onOff && getgdesc(GD_STEREO)) {
    MonoMonitor = getmonitor();
    setmonitor(STR_RECT);
    StereoState = TRUE;
  } else {
    setmonitor(MonoMonitor);
    StereoState = FALSE;
  }

  StereoSwitch = onOff;
}

void
SgGLSetEye(SgEye eye)
{
  Screencoord xMin, xMax;

  xMin = VpX + 0.5;
  xMax = VpX + VpW + 0.5;

  if (eye == SG_EYE_LEFT)
    viewport(xMin, xMax, YOFFSET, YOFFSET + YMAXSTEREO);
  else
    viewport(xMin, xMax, 0, YMAXSTEREO);
}

void
SgGLConvCoord(int xs, int ys, float *xP, float *yP)
{
  *xP = (2 * (xs - VpX) - VpW) / VpH;
  if (StereoState) {
    if (ys >= 0.5 * VpH)
      *yP = 4 * (ys - VpY) / VpH - 3.0;
    else
      *yP = 4 * (ys - VpY) / VpH - 1.0;
  } else {
    *yP = 2 * (ys - VpY) / VpH - 1.0;
  }
}

void
SgGLUpdateView(void)
{
  float b;

  mmode(MPROJECTION);
  if (Proj == SG_PROJ_ORTHO) {
    b = SgGetOrthoHeight();
    ortho(- b * AspRat, b * AspRat, - b, b, Near, Far);
  } else {
    perspective(ViewAng, AspRat, Near, Far);
  }
  mmode(MVIEWING);

  loadmatrix(IdentM);
  lookat(VX, VY, VZ, 0.0, 0.0, 0.0, 0.0);
}

void
SgGLSetProjection(SgProjection proj)
{
  Proj = proj;
}

void
SgGLSetViewPoint(float x[3])
{
  VX = x[0];
  VY = x[1];
  VZ = x[2];
}

void
SgGLSetViewAngle(float ang)
{
  /* argument is in rad, GL angles are in tenths of degrees */
  ViewAng = Rad2GLAng(ang);
}

void
SgGLSetNearPlane(float d)
{
  Near = d;
}

void
SgGLSetFarPlane(float d)
{
  Far = d;
}

static void
setViewport(void)
{
  viewport((Screencoord) (VpX + 0.5), (Screencoord) (VpX + VpW + 0.5),
      (Screencoord) (VpY + 0.5), (Screencoord) (VpY + VpH + 0.5));
}

void
SgGLSetViewport(float x, float y, float width, float height)
{
  VpX = x;
  VpY = y;
  VpW = width;
  VpH = height;
  AspRat = width / height;

  setViewport();
}

void
SgGLSetBuffer(SgBuffer buf)
{
  if (buf == SG_BUF_FRONT) {
    backbuffer(FALSE);
    frontbuffer(TRUE);
  } else {
    frontbuffer(FALSE);
    backbuffer(TRUE);
  }
}

void
SgGLEndFrame(void)
{
  setViewport();  /* may have been changed for stereo display */
  swapbuffers();
}

void
GLStart2D(void)
{
  pushmatrix();
  loadmatrix(IdentM);
  mmode(MPROJECTION);
  getmatrix(ProjM);
  ortho2(- AspRat, AspRat, -1.0, 1.0);
  zbuffer(FALSE);
}

void
GLEnd2D(void)
{
  loadmatrix(ProjM);
  mmode(MVIEWING);
  popmatrix();
  if (SgGetFeature(SG_FEATURE_HIDDEN))
    zbuffer(TRUE);
}
